#!/usr/bin/env python3

from binascii import hexlify, unhexlify
from mnemonic import Mnemonic

import hmac
import hashlib

from ..libs.segwit import encode, decode
from ..libs.ed25519 import (
    encodeint, encodepoint, decodeint, decodepoint,
    scalarmultbase, edwards_double, edwards_add
)
from ..signature import sign, verify
from ..utils import get_mnemonic_language, is_mnemonic
from ..rpc import config, account_create, get_balance
from .tools import get_xpublic_key
from .utils import (
    prune_root_scalar, prune_intermediate_scalar,
    get_bytes, bad_seed_checker
)


class Wallet:
    """
    Bytom wallet class.

    :param network: Bytom network, defaults to solonet.
    :type network: str
    :returns:  Wallet -- Bytom wallet instance.

    .. note::
        Bytom has only three networks, ``mainnet``, ``solonet`` and ``testnet``.
    """

    def __init__(self, network="solonet"):

        # Bytom network.
        if not isinstance(network, str):
            raise TypeError("network must be string format")
        if network not in "mainnet/solonet/testnet".split("/"):
            raise ValueError("invalid network option, choose only mainnet, solonet and testnet network")
        self.network = network

        self._entropy, self._mnemonic, self._passphrase, \
            self._language, self._seed = None, None, None, None, None
        self._xprivate_key, self._xpublic_key, self._indexes, \
            self._path, self._guid = None, None, list(), None, None

    def from_entropy(self, entropy, passphrase=None, language="english"):
        """
        Initialize Bytom wallet from entropy.

        :param entropy: Entropy hex string.
        :type entropy: str.
        :param passphrase: Secret password/passphrase, default to None.
        :type passphrase: str.
        :param language: Mnemonic language, default to english.
        :type language: str.
        :returns:  Wallet -- Bytom wallet class instance.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_entropy("8d7454bb99e8e68de6adfc5519cbee64")
        <pybytom.wallet.Wallet object at 0x040DA268>
        """

        if passphrase is None:
            passphrase = str()

        # Checking parameters
        if not isinstance(entropy, str):
            raise TypeError("entropy must be string format")
        if not isinstance(passphrase, str):
            raise TypeError("passphrase must be string format")
        if not isinstance(language, str):
            raise TypeError("language must be string format")
        if language not in "english/french/italian/spanish/" \
                           "chinese_simplified/chinese_traditional/japanese/korean".split("/"):
            raise ValueError("invalid language option, choose only english, french, italian, spanish, "
                             "chinese_simplified, chinese_traditional, japanese & korean language")

        self._entropy = unhexlify(entropy.encode())
        self._passphrase, self._language = passphrase, language
        self._mnemonic = Mnemonic(language=self._language) \
            .to_mnemonic(data=self._entropy)
        self._seed = Mnemonic.to_seed(
            mnemonic=self._mnemonic, passphrase=self._passphrase)
        self.from_seed(seed=hexlify(self._seed).decode())
        return self

    def from_mnemonic(self, mnemonic, passphrase=None, language=None):
        """
        Initialize Bytom wallet from mnemonic.

        :param mnemonic: 12 words mnemonic.
        :type mnemonic: str.
        :param passphrase: Secret password/passphrase, default to None.
        :type passphrase: str.
        :param language: Mnemonic language, default to None.
        :type language: str.
        :returns:  Wallet -- Bytom wallet class instance.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_mnemonic("舒 使 系 款 株 擾 麼 鄉 狗 振 誤 謀")
        <pybytom.wallet.Wallet object at 0x040DA268>
        """

        if passphrase is None:
            passphrase = str()

        # Checking parameters
        if not isinstance(mnemonic, str):
            raise TypeError("mnemonic must be string format")
        if not isinstance(passphrase, str):
            raise TypeError("passphrase must be string format")
        if language and language not in "english/french/italian/spanish/chinese_simplified" \
                                        "/chinese_traditional/japanese/korean".split("/"):
            raise ValueError("invalid language option, choose only english, french, italian, spanish, "
                             "chinese_simplified, chinese_traditional, japanese & korean language")
        if not is_mnemonic(mnemonic=mnemonic, language=language):
            raise ValueError("invalid 12 word mnemonic.")

        self._mnemonic = mnemonic
        self._passphrase, self._language = \
            passphrase, get_mnemonic_language(mnemonic=self._mnemonic)
        self._seed = Mnemonic.to_seed(
            mnemonic=self._mnemonic, passphrase=self._passphrase)
        self.from_seed(seed=hexlify(self._seed).decode())
        return self

    def from_seed(self, seed):
        """
        Initialize Bytom wallet from seed.

        :param seed: Mnemonic seed hex string.
        :type seed: str.
        :returns:  Wallet -- Bytom wallet class instance.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_seed("baff3e1fe60e1f2a2d840d304acc98d1818140c79354a353b400fb019bfb256bc392d7aa9047adff1f14bce0342e14605c6743a6c08e02150588375eb2eb7d49")
        <pybytom.wallet.Wallet object at 0x040DA268>
        """

        # Checking parameters
        if not isinstance(seed, str):
            raise TypeError("seed must be string format")

        self._seed = unhexlify(seed.encode())
        i = hmac.HMAC(b"Root", get_bytes(self._seed),
                      digestmod=hashlib.sha512).hexdigest()
        il, ir = i[:64], i[64:]
        bad_seed_checker(il, True)

        # get root xprivate_key key
        self._xprivate_key = prune_root_scalar(il).hex() + ir
        return self

    def from_xprivate_key(self, xprivate_key):
        """
        Initialize Bytom wallet from seed.

        :param xprivate_key: Bytom xprivate key.
        :type xprivate_key: str.
        :returns:  Wallet -- Bytom wallet class instance.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_xprivate_key("205b15f70e253399da90b127b074ea02904594be9d54678207872ec1ba31ee51ef4490504bd2b6f997113671892458830de09518e6bd5958d5d5dd97624cfa4b")
        <pybytom.wallet.Wallet object at 0x040DA268>
        """

        # Checking parameters
        if not isinstance(xprivate_key, str):
            raise TypeError("xprivate key must be string format")

        self._xprivate_key = xprivate_key
        return self

    def derive_private_key(self, index):
        index = int(index).to_bytes(4, byteorder="little").hex()
        self._indexes.append(index)
        return self

    def from_indexes(self, indexes):
        """
        Drive Bytom wallet from indexes.

        :param indexes: Bytom derivation indexes.
        :type indexes: list.
        :returns:  Wallet -- Bytom wallet class instance.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_xprivate_key("205b15f70e253399da90b127b074ea02904594be9d54678207872ec1ba31ee51ef4490504bd2b6f997113671892458830de09518e6bd5958d5d5dd97624cfa4b")
        >>> wallet.from_indexes(["2c000000", "99000000", "01000000", "00000000", "01000000"])
        <pybytom.wallet.Wallet object at 0x040DA268>
        """

        # Checking parameters
        if not isinstance(indexes, list):
            raise TypeError("indexes must be list format")

        self._indexes = indexes
        return self

    def from_index(self, index, harden=False):
        """
        Drive Bytom wallet from index.

        :param index: Bytom derivation index.
        :type index: int.
        :param harden: BIP 32 key harden, default to False.
        :type harden: bool.
        :returns:  Wallet -- Bytom wallet class instance.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_entropy(entropy="8d7454bb99e8e68de6adfc5519cbee64", language="japaneses")
        >>> wallet.from_index(44)
        >>> wallet.from_index(153)
        >>> wallet.from_index(1)  # Account
        >>> wallet.from_index(0)  # Change False(0)/True(1)
        >>> wallet.from_index(1)  # Address
        <pybytom.wallet.Wallet object at 0x040DA268>
        """

        # Checking parameters
        if not isinstance(index, int):
            raise TypeError("index must be integer format")

        if harden:
            self.derive_private_key(index + 0x80000000)
        else:
            self.derive_private_key(index)
        return self

    def from_path(self, path):
        """
        Drive Bytom wallet from path.

        :param path: Bytom derivation path.
        :type path: str.
        :returns:  Wallet -- Bytom wallet instance.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_mnemonic(mnemonic="舒 使 系 款 株 擾 麼 鄉 狗 振 誤 謀", passphrase="Hello Meheret!")
        >>> wallet.from_path("m/44/153/1/0/1")
        <pybytom.wallet.Wallet object at 0x040DA268>
        """

        # Checking parameters
        if not isinstance(path, str):
            raise TypeError("path must be string format")
        if str(path)[0:2] != "m/":
            raise ValueError("bad path, insert like this type of path \"m/0'/0\"! ")

        for index in path.lstrip("m/").split("/"):
            if "'" in index:
                self.derive_private_key(int(index[:-1]) + 0x80000000)
            else:
                self.derive_private_key(int(index))
        return self

    def clean_derivation(self):
        """
        Clean derivation indexes/path.

        :returns:  Wallet -- Bytom wallet instance.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_mnemonic(mnemonic="舒 使 系 款 株 擾 麼 鄉 狗 振 誤 謀", passphrase="Hello Meheret!")
        >>> wallet.from_path("m/44/153/1/0/1")
        >>> wallet.indexes()
        ["2c000000", "99000000", "01000000", "00000000", "01000000"]
        >>> wallet.path()
        "m/44/153/1/0/1"
        >>> wallet.clean_derivation()
        >>> wallet.indexes()
        []
        >>> wallet.path()
        None
        """
        self._indexes = list()
        return self

    # Getting guid from blockcenter
    def guid(self):
        """
        Get Bytom wallet blockcenter guid.

        :return: str -- Bytom blockcenter guid.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_mnemonic("indicate warm sock mistake code spot acid ribbon sing over taxi toast")
        >>> wallet.guid()
        "f0ed6ddd-9d6b-49fd-8866-a52d1083a13b"
        """

        if self._guid is None:
            self._guid = account_create(
                xpublic_key=self.xpublic_key(), network=self.network)["guid"]
        return self._guid

    def entropy(self):
        """
        Get wallet entropy.

        :return: str -- Entropy hex string.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_entropy(entropy="8d7454bb99e8e68de6adfc5519cbee64", language="korean")
        >>> wallet.entropy()
        "8d7454bb99e8e68de6adfc5519cbee64"
        """

        return hexlify(self._entropy).decode() if self._entropy else None

    def mnemonic(self):
        """
        Get wallet mnemonic.

        :return: str -- 12 word mnemonic.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_entropy(entropy="8d7454bb99e8e68de6adfc5519cbee64", language="italian")
        >>> wallet.mnemonic()
        "occasione pizzico coltivato cremoso odorare epilogo patacca salone fonia sfuso vispo selettivo"
        """

        return str(self._mnemonic) if self._mnemonic else None

    def passphrase(self):
        """
        Get wallet secret password/passphrase.

        :return: str -- Secret password/passphrase.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_entropy(entropy="8d7454bb99e8e68de6adfc5519cbee64", passphrase="Hello Meheret!", language="italian")
        >>> wallet.passphrase()
        "Hello Meheret!"
        """

        return str(self._passphrase) if self._passphrase else None

    def language(self):
        """
        Get wallet mnemonic language.

        :return: str -- Mnemonic language.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_entropy(entropy="8d7454bb99e8e68de6adfc5519cbee64", language="italian")
        >>> wallet.language()
        "italian"
        """

        return str(self._language) if self._language else None

    def seed(self):
        """
        Get wallet mnemonic seed.

        :return: str -- Mnemonic seed.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_mnemonic("indicate warm sock mistake code spot acid ribbon sing over taxi toast")
        >>> wallet.seed()
        "baff3e1fe60e1f2a2d840d304acc98d1818140c79354a353b400fb019bfb256bc392d7aa9047adff1f14bce0342e14605c6743a6c08e02150588375eb2eb7d49"
        """

        return hexlify(self._seed).decode() if self._seed else None

    def xprivate_key(self):
        """
        Get Bytom wallet xprivate key.

        :return: str -- Bytom xprivate key.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_mnemonic("indicate warm sock mistake code spot acid ribbon sing over taxi toast")
        >>> wallet.xprivate_key()
        "205b15f70e253399da90b127b074ea02904594be9d54678207872ec1ba31ee51ef4490504bd2b6f997113671892458830de09518e6bd5958d5d5dd97624cfa4b"
        """

        return str(self._xprivate_key)

    def xpublic_key(self):
        """
        Get Bytom wallet xpublic key.

        :return: str -- Bytom xpublic key.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_mnemonic("indicate warm sock mistake code spot acid ribbon sing over taxi toast")
        >>> wallet.xpublic_key()
        "16476b7fd68ca2acd92cfc38fa353e75d6103f828276f44d587e660a6bd7a5c5ef4490504bd2b6f997113671892458830de09518e6bd5958d5d5dd97624cfa4b"
        """

        xprivate_bytes = get_bytes(self.xprivate_key())
        scalar = decodeint(xprivate_bytes[:len(xprivate_bytes) // 2])
        buf = encodepoint(scalarmultbase(scalar))
        xpublic_key = buf + xprivate_bytes[len(xprivate_bytes) // 2:]
        self._xpublic_key = xpublic_key.hex()
        return self._xpublic_key

    def expand_xprivate_key(self):
        """
        Get Bytom wallet expand xprivate key.

        :return: str -- Bytom expand xprivate key.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_mnemonic("indicate warm sock mistake code spot acid ribbon sing over taxi toast")
        >>> wallet.expand_xprivate_key()
        "205b15f70e253399da90b127b074ea02904594be9d54678207872ec1ba31ee5102416c643cfb46ab1ae5a524c8b4aaa002eb771d0d9cfc7490c0c3a8177e053e"
        """

        i = hmac.HMAC(b"Expand", get_bytes(self._xprivate_key),
                      digestmod=hashlib.sha512).hexdigest()
        il, ir = i[:64], i[64:]
        bad_seed_checker(ir, True)

        expand_xprivate = self._xprivate_key[:64] + ir
        return expand_xprivate

    def private_key(self):
        """
        Get Bytom wallet private key.

        :return: str -- Bytom private key.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_mnemonic("indicate warm sock mistake code spot acid ribbon sing over taxi toast")
        >>> wallet.from_path("m/44/153/1/0/1")
        >>> wallet.private_key()
        "e07af52746e7cccd0a7d1fba6651a6f474bada481f34b1c5bab5e2d71e36ee515803ee0a6682fb19e279d8f4f7acebee8abd0fc74771c71565f9a9643fd77141"
        """

        return self.child_xprivate_key()

    def public_key(self):
        """
        Get Bytom wallet public key.

        :return: str -- Bytom public key.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_mnemonic("indicate warm sock mistake code spot acid ribbon sing over taxi toast")
        >>> wallet.from_path("m/44/153/1/0/1")
        >>> wallet.public_key()
        "91ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e2"
        """

        return self.child_xpublic_key()[:64]

    def indexes(self):
        """
        Get Bytom wallet derivation indexes.

        :return: list -- Bytom derivation indexes.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_mnemonic("indicate warm sock mistake code spot acid ribbon sing over taxi toast")
        >>> wallet.from_path("m/44/153/1/0/1")
        >>> wallet.indexes()
        ["2c000000", "99000000", "01000000", "00000000", "01000000"]
        """

        return self._indexes

    def path(self):
        """
        Get Bytom wallet derivation path.

        :return: str -- Bytom derivation path.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_mnemonic("indicate warm sock mistake code spot acid ribbon sing over taxi toast")
        >>> wallet.from_indexes(["2c000000", "99000000", "01000000", "00000000", "01000000"])
        >>> wallet.path()
        "m/44/153/1/0/1"
        """

        path = "m/"
        for i, index in enumerate(self._indexes, 1):
            number = int.from_bytes(bytes.fromhex(index), byteorder="little")
            if i == len(self._indexes):
                path = path + str(number)
            else:
                path = path + str(number) + "/"
        return path if not path == "m/" else None

    def child_xprivate_key(self):
        """
        Get Bytom get child xprivate key.

        :return: str -- Bytom child xprivate key.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_mnemonic("indicate warm sock mistake code spot acid ribbon sing over taxi toast")
        >>> wallet.from_path("m/44/153/1/0/1")
        >>> wallet.child_xprivate_key()
        "e07af52746e7cccd0a7d1fba6651a6f474bada481f34b1c5bab5e2d71e36ee515803ee0a6682fb19e279d8f4f7acebee8abd0fc74771c71565f9a9643fd77141"
        """

        xprivate_key = self.xprivate_key()
        for index in range(len(self._indexes)):
            index_bytes = get_bytes(self._indexes[index])
            xpublic_key = get_xpublic_key(xprivate_key=xprivate_key)
            xpublic_bytes = get_bytes(xpublic_key)
            xprivate_bytes = get_bytes(xprivate_key)
            i = bytearray(hmac.HMAC(xpublic_bytes[32:],
                                    b"N" + xpublic_bytes[:32] + index_bytes,
                                    digestmod=hashlib.sha512).digest())
            il, ir = i[:32], i[32:]
            bad_seed_checker(il)

            i = prune_intermediate_scalar(il)[:32] + ir

            carry = 0
            total = 0
            for _i in range(32):
                total = xprivate_bytes[_i] + i[_i] + carry
                i[_i] = total & 0xff
                carry = total >> 8
            if (total >> 8) != 0:
                print("sum does not fit in 256-bit int")
            xprivate_key = i.hex()

        child_xprivate = xprivate_key
        return child_xprivate

    def child_xpublic_key(self):
        """
        Get Bytom get child xpublic key.

        :return: str -- Bytom child xpublic key.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_mnemonic("indicate warm sock mistake code spot acid ribbon sing over taxi toast")
        >>> wallet.from_path("m/44/153/1/0/1")
        >>> wallet.child_xpublic_key()
        "91ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e25803ee0a6682fb19e279d8f4f7acebee8abd0fc74771c71565f9a9643fd77141"
        """

        xpublic_key = self.xpublic_key()
        for index in range(len(self._indexes)):
            index_bytes = get_bytes(self._indexes[index])
            xpublic_bytes = get_bytes(xpublic_key)
            i = bytearray(hmac.HMAC(xpublic_bytes[32:],
                                    b"N" + xpublic_bytes[:32] + index_bytes,
                                    digestmod=hashlib.sha512).digest())

            il, ir = i[:32], i[32:]
            bad_seed_checker(il)

            f = bytes(prune_intermediate_scalar(il))
            scalar = decodeint(f)
            f = scalarmultbase(scalar)

            p = decodepoint(xpublic_bytes[:32])
            p = edwards_add(p, f)
            public_key = encodepoint(p)

            xpublic_bytes = public_key[:32] + ir
            xpublic_key = xpublic_bytes.hex()

        child_xpublic = xpublic_key
        return child_xpublic

    def program(self):
        """
        Get Bytom wallet control program.

        :return: str -- Bytom wallet control program.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_mnemonic("indicate warm sock mistake code spot acid ribbon sing over taxi toast")
        >>> wallet.from_indexes(["2c000000", "99000000", "01000000", "00000000", "01000000"])
        >>> wallet.program()
        "00142cda4f99ea8112e6fa61cdd26157ed6dc408332a"
        """

        public = self.public_key()
        public_byte = get_bytes(public)

        ripemd160 = hashlib.new("ripemd160")
        ripemd160.update(public_byte)
        public_hash = ripemd160.hexdigest()
        program = "0014" + public_hash
        return program

    def address(self, network=None):
        """
        Get Bytom wallet address.

        :param network: Bytom network, defaults to solonet.
        :type network: str
        :return: str -- Bytom wallet address.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_mnemonic("indicate warm sock mistake code spot acid ribbon sing over taxi toast")
        >>> wallet.from_indexes(["2c000000", "99000000", "01000000", "00000000", "01000000"])
        >>> wallet.address()
        "bm1q9ndylx02syfwd7npehfxz4lddhzqsve2fu6vc7"
        """

        if network is None:
            network = self.network
        if not isinstance(network, str):
            raise TypeError("network must be string format")

        if network not in "mainnet/solonet/testnet".split("/"):
            raise ValueError("invalid network option, choose only mainnet, solonet and testnet network")
        elif network == "mainnet":
            return encode("bm", 0, get_bytes(self.program()[4:]))
        elif network == "solonet":
            return encode("sm", 0, get_bytes(self.program()[4:]))
        elif network == "testnet":
            return encode("tm", 0, get_bytes(self.program()[4:]))

    def balance(self, asset=config["BTM_ASSET"]):
        """
        Get Bytom wallet balance.

        :param asset: Bytom asset, defaults to BTM asset.
        :type asset: str
        :return: int -- Bytom balance.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet(network="mainnet")
        >>> wallet.from_mnemonic("indicate warm sock mistake code spot acid ribbon sing over taxi toast")
        >>> wallet.balance()
        2450000000
        """

        return get_balance(address=self.address(), asset=asset, network=self.network)

    def sign(self, message):
        """
        Sign message data by private key.

        :param message: Message data.
        :type message: str.
        :return: str -- Signed message data (signature).

        >>> from pybytom.wallet import Wallet
        >>> message = "1246b84985e1ab5f83f4ec2bdf271114666fd3d9e24d12981a3c861b9ed523c6"
        >>> wallet = Wallet(network="testnet")
        >>> wallet.from_mnemonic("indicate warm sock mistake code spot acid ribbon sing over taxi toast")
        >>> wallet.from_indexes(["2c000000", "99000000", "01000000", "00000000", "01000000"])
        >>> wallet.sign(message=message)
        "f6624fea84fadccbc1bc72dc384f662468e271c4e32d846bc0a1524470549992c8ffcc3ca43891a30de4235392b0868c506ed254f0f77cc1f2b9c1a2385ddb05"
        """

        if not isinstance(message, str):
            raise TypeError("message must be string format")

        return sign(self.private_key(), message)

    def verify(self, message, signature):
        """
        Verify signature by public key.

        :param message: Message data.
        :type message: str.
        :param signature: Signed message data.
        :type signature: str.
        :return: bool -- Verified signature (True/False).

        >>> from pybytom.wallet import Wallet
        >>> message = "1246b84985e1ab5f83f4ec2bdf271114666fd3d9e24d12981a3c861b9ed523c6"
        >>> signature = "f6624fea84fadccbc1bc72dc384f662468e271c4e32d846bc0a1524470549992c8ffcc3ca43891a30de4235392b0868c506ed254f0f77cc1f2b9c1a2385ddb05"
        >>> wallet = Wallet(network="testnet")
        >>> wallet.from_mnemonic("indicate warm sock mistake code spot acid ribbon sing over taxi toast")
        >>> wallet.from_indexes(["2c000000", "99000000", "01000000", "00000000", "01000000"])
        >>> wallet.verify(message=message, signature=signature)
        True
        """

        if not isinstance(message, str):
            raise TypeError("message must be string format")
        if not isinstance(signature, str):
            raise TypeError("signature must be string format")

        return verify(self.public_key(), message, signature)

    def dumps(self):
        """
        Get Bytom all wallet information's

        :return: dict -- Bytom all wallet information's.

        >>> from pybytom.wallet import Wallet
        >>> wallet = Wallet()
        >>> wallet.from_mnemonic("indicate warm sock mistake code spot acid ribbon sing over taxi toast")
        >>> wallet.from_indexes(["2c000000", "99000000", "01000000", "00000000", "01000000"])
        >>> wallet.dumps()
        {'entropy': None, 'mnemonic': 'indicate warm sock mistake code spot acid ribbon sing over taxi toast', 'language': 'english', 'passphrase': None, 'seed': 'baff3e1fe60e1f2a2d840d304acc98d1818140c79354a353b400fb019bfb256bc392d7aa9047adff1f14bce0342e14605c6743a6c08e02150588375eb2eb7d49', 'xprivate_key': '205b15f70e253399da90b127b074ea02904594be9d54678207872ec1ba31ee51ef4490504bd2b6f997113671892458830de09518e6bd5958d5d5dd97624cfa4b', 'xpublic_key': '16476b7fd68ca2acd92cfc38fa353e75d6103f828276f44d587e660a6bd7a5c5ef4490504bd2b6f997113671892458830de09518e6bd5958d5d5dd97624cfa4b', 'expand_xprivate_key': '205b15f70e253399da90b127b074ea02904594be9d54678207872ec1ba31ee5102416c643cfb46ab1ae5a524c8b4aaa002eb771d0d9cfc7490c0c3a8177e053e', 'indexes': ['2c000000', '99000000', '01000000', '00000000', '01000000'], 'path': 'm/44/153/1/0/1', 'child_xprivate_key': 'e07af52746e7cccd0a7d1fba6651a6f474bada481f34b1c5bab5e2d71e36ee515803ee0a6682fb19e279d8f4f7acebee8abd0fc74771c71565f9a9643fd77141', 'child_xpublic_key': '91ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e25803ee0a6682fb19e279d8f4f7acebee8abd0fc74771c71565f9a9643fd77141', 'private_key': 'e07af52746e7cccd0a7d1fba6651a6f474bada481f34b1c5bab5e2d71e36ee515803ee0a6682fb19e279d8f4f7acebee8abd0fc74771c71565f9a9643fd77141', 'public_key': '91ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e2', 'program': '00142cda4f99ea8112e6fa61cdd26157ed6dc408332a', 'address': {'mainnet': 'bm1q9ndylx02syfwd7npehfxz4lddhzqsve2fu6vc7', 'solonet': 'sm1q9ndylx02syfwd7npehfxz4lddhzqsve2gdsdcs', 'testnet': 'tm1q9ndylx02syfwd7npehfxz4lddhzqsve2d2mgc0'}}
        """

        return dict(
            entropy=self.entropy(),
            mnemonic=self.mnemonic(),
            language=self.language(),
            passphrase=self.passphrase(),
            seed=self.seed(),
            xprivate_key=self.xprivate_key(),
            xpublic_key=self.xpublic_key(),
            expand_xprivate_key=self.expand_xprivate_key(),
            indexes=self.indexes(),
            path=self.path(),
            child_xprivate_key=self.child_xprivate_key(),
            child_xpublic_key=self.child_xpublic_key(),
            private_key=self.private_key(),
            public_key=self.public_key(),
            program=self.program(),
            address=dict(
                mainnet=self.address(network="mainnet"),
                solonet=self.address(network="solonet"),
                testnet=self.address(network="testnet")
            )
        )



#!/usr/bin/env python3

from .wallet import Wallet
from .tools import get_xpublic_key, get_expand_xprivate_key, get_child_xprivate_key, indexes_to_path, \
    get_child_xpublic_key, get_address, get_program, get_private_key, get_public_key, get_bytes, path_to_indexes


# Constant values
BIP32KEY_HARDEN = 0x80000000
# Derivation Path
PATH = "m/44/153/1/0/1"
# Derivation Indexes
INDEXES = [
    "2c000000",  # 44
    "99000000",  # 153
    "01000000",  # 1 Account
    "00000000",  # 0 Change
    "01000000"  # 1 Address
]


__all__ = [
    "Wallet",
    "get_xpublic_key",
    "get_expand_xprivate_key",
    "get_child_xprivate_key",
    "get_child_xpublic_key",
    "get_address",
    "get_program",
    "get_private_key",
    "get_public_key",
    "get_bytes",
    "indexes_to_path",
    "path_to_indexes"
]



#!/usr/bin/env python3

from ..rpc import config
from ..wallet import Wallet
from ..wallet.tools import (
    indexes_to_path, get_program, get_address
)
from ..rpc import build_transaction
from .tools import spend_wallet_action, control_address_action
from ..exceptions import ClientError, NetworkError


# Byrom transaction
class Transaction:
    """
    Bytom Transaction class.

    :param network: Bytom network, defaults to solonet.
    :type network: str

    :returns: Transaction -- Bytom transaction instance.
    .. note::
        Bytom has only three networks, ``mainnet``. ``solonet`` and ``testnet``.
    """

    # Initialization transaction
    def __init__(self, network=config["network"]):

        # Checking network
        if not isinstance(network, str):
            raise TypeError("invalid network instance, only takes string type")
        if network not in ["mainnet", "solonet", "testnet"]:
            raise NetworkError("invalid network type",
                               "choose only mainnet, solonet or testnet networks")
        # Bytom network and transaction
        self.network, self.transaction = network, None
        # Bytom fee, signed datas/signatures and confirmations
        self._fee, self._signatures, self._confirmations = \
            config["fee"], [], config["confirmations"]

    # Building bytom transaction
    def build_transaction(self, **kwargs):
        """
        Bytom build transaction.

        :param kwargs: Arbitrary keyword arguments. If you do accept ``**kwargs``, make sure
        you link to documentation that describes what keywords are accepted,
        or list the keyword arguments as a definition list:
        ``guid`` (str) bytom blockcenter guid.
        ``inputs`` (list) bytom blockcenter inputs.
        ``outputs`` (list) bytom blockcenter outputs.
        ``fee`` (int) bytom transaction fee, defaults to 10000000.
        ``confirmations`` (int) confirmations: bytom transaction confirmations, defaults to 1.
        :returns:  Transaction -- Bytom transaction instance.

        >>> from pybytom.transaction import Transaction
        >>> transaction = Transaction(network="mainnet")
        >>> transaction.build_transaction(guid="f0ed6ddd-9d6b-49fd-8866-a52d1083a13b", inputs=[[...], ...], outputs=[[...], ...], fee=10000000, confirmations=3)
        <pybytom.transaction.transaction.Transaction object at 0x0409DAF0>
        """

        # Checking build transaction arguments instance
        if {"guid", "inputs", "outputs"} - set(kwargs.keys()):
            raise ClientError("you can't build transaction without 'guid', 'inputs' and 'outputs'",
                              "default fee is 10000000 and confirmations to 1.")
        if not isinstance(kwargs["guid"], str):
            raise TypeError("invalid guid instance, only takes string type")
        if not isinstance(kwargs["inputs"], list):
            raise TypeError("invalid inputs instance, only takes list type")
        if not isinstance(kwargs["outputs"], list):
            raise TypeError("invalid outputs instance, only takes list type")
        if "fee" in kwargs.keys() and not isinstance(kwargs["fee"], int):
            raise TypeError("invalid fee instance, only takes integer type")
        if "confirmations" in kwargs.keys() and not isinstance(kwargs["confirmations"], int):
            raise TypeError("invalid confirmations instance, only takes integer type")

        # Setting fee and confirmations
        self._fee = kwargs["fee"] \
            if "fee" in kwargs.keys() else config["fee"]
        self._confirmations = kwargs["confirmations"] \
            if "confirmations" in kwargs.keys() else config["confirmations"]

        # Transaction config
        transaction = dict(
            guid=kwargs["guid"],
            inputs=kwargs["inputs"],
            outputs=kwargs["outputs"],
            fee=self._fee,
            confirmations=self._confirmations
        )
        # Building transaction
        self.transaction = build_transaction(
            transaction=transaction, network=self.network)
        return self

    # Transaction fee
    def fee(self) -> int:
        """
        Bytom transaction fee.

        :returns: list -- Bytom transaction fee.

        >>> from pybytom.transaction import Transaction
        >>> transaction = Transaction(network="mainnet")
        >>> transaction.build_transaction("f0ed6ddd-9d6b-49fd-8866-a52d1083a13b", inputs=[...], outputs=[...])
        >>> transaction.fee()
        10000000
        """

        return self._fee

    # Transaction confirmations
    def confirmations(self) -> int:
        """
        Bytom transaction confirmations.

        :returns: int -- Bytom transaction confirmations.

        >>> from pybytom.transaction import Transaction
        >>> transaction = Transaction(network="mainnet")
        >>> transaction.build_transaction("f0ed6ddd-9d6b-49fd-8866-a52d1083a13b", inputs=[...], outputs=[...])
        >>> transaction.confirmations()
        2
        """

        return self._confirmations

    # Transaction hash
    def hash(self) -> str:
        """
        Bytom transaction hash.

        :returns: str -- Bytom transaction hash or transaction id.

        >>> from pybytom.transaction import Transaction
        >>> transaction = Transaction(network="mainnet")
        >>> transaction.build_transaction("f0ed6ddd-9d6b-49fd-8866-a52d1083a13b", inputs=[...], outputs=[...])
        >>> transaction.hash()
        "2993414225f65390220730d0c1a356c14e91bca76db112d37366df93e364a492"
        """

        if self.transaction is None:
            raise ValueError("transaction is none, build transaction first.")
        return self.transaction["tx"]["hash"]

    # Transaction json
    def json(self):
        """
        Bytom transaction json format.

        :returns: dict -- Bytom transaction json format.

        >>> from pybytom.transaction import Transaction
        >>> transaction = Transaction(network="mainnet")
        >>> transaction.build_transaction("f0ed6ddd-9d6b-49fd-8866-a52d1083a13b", inputs=[...], outputs=[...])
        >>> transaction.json()
        {"hash": "2993414225f65390220730d0c1a356c14e91bca76db112d37366df93e364a492", "status_fail": false, "size": 379, "submission_timestamp": 0, "memo": "", "inputs": [{"script": "00142cda4f99ea8112e6fa61cdd26157ed6dc408332a", "address": "bm1q9ndylx02syfwd7npehfxz4lddhzqsve2fu6vc7", "asset": "ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff", "amount": 2450000000, "type": "spend"}], "outputs": [{"utxo_id": "5edccebe497893c289121f9e365fdeb34c97008b9eb5a9960fe9541e7923aabc", "script": "01642091ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e220ac13c0bb1445423a641754182d53f0677cd4351a0e743e6f10b35122c3d7ea01202b9a5949f5546f63a253e41cda6bffdedb527288a7e24ed953f5c2680c70d6ff741f547a6416000000557aa888537a7cae7cac631f000000537acd9f6972ae7cac00c0", "address": "smart contract", "asset": "ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff", "amount": 1000, "type": "control"}, {"utxo_id": "f8cfbb692db1963be88b09c314adcc9e19d91c6c019aa556fb7cb76ba8ffa1fa", "script": "00142cda4f99ea8112e6fa61cdd26157ed6dc408332a", "address": "bm1q9ndylx02syfwd7npehfxz4lddhzqsve2fu6vc7", "asset": "ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff", "amount": 2439999000, "type": "control"}], "fee": 10000000, "balances": [{"asset": "ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff", "amount": "-10001000"}], "types": ["ordinary"]}
        """
        if self.transaction is None:
            raise ValueError("transaction is none, build transaction first.")
        return self.transaction["tx"]

    # Transaction raw
    def raw(self) -> str:
        """
        Bytom transaction raw.

        :returns: str -- Bytom transaction raw.

        >>> from pybytom.transaction import Transaction
        >>> transaction = Transaction(network="mainnet")
        >>> transaction.build_transaction("f0ed6ddd-9d6b-49fd-8866-a52d1083a13b", inputs=[...], outputs=[...])
        >>> transaction.raw()
        "070100010160015e7f2d7ecec3f61d30d0b2968973a3ac8448f0599ea20dce883b48c903c4d6e87fffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff8091a0900901011600142cda4f99ea8112e6fa61cdd26157ed6dc408332a22012091ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e20201ad01ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffe80701880101642091ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e220ac13c0bb1445423a641754182d53f0677cd4351a0e743e6f10b35122c3d7ea01202b9a5949f5546f63a253e41cda6bffdedb527288a7e24ed953f5c2680c70d6ff741f547a6416000000557aa888537a7cae7cac631f000000537acd9f6972ae7cac00c000013dffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff98dcbd8b09011600142cda4f99ea8112e6fa61cdd26157ed6dc408332a00"
        """

        if self.transaction is None:
            raise ValueError("transaction is none, build transaction first.")
        return self.transaction["raw_transaction"]

    # Unsigned datas with instructions
    def unsigned_datas(self, detail=False) -> list:
        """
        Bytom transaction json format.

        :returns: dict -- Bytom transaction json format.

        >>> from pybytom.transaction import Transaction
        >>> transaction = Transaction(network="mainnet")
        >>> transaction.build_transaction("f0ed6ddd-9d6b-49fd-8866-a52d1083a13b", inputs=[...], outputs=[...])
        >>> transaction.unsigned_datas()
        [{'datas': ['a7ba7ffd969e4150b36ea9f1f861d53c6f69226a3519023899d27bf1f85be3d9'], 'public_key': '91ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e2', 'network': 'mainnet', 'path': 'm/44/153/1/0/1'}, {'datas': ['0dbf0e408099017711a941d9a9899541afa64eb31a7118815d37e6d603b7c329'], 'public_key': '91ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e2', 'network': 'mainnet', 'path': 'm/44/153/1/0/1'}]
        """

        # Checking transaction
        if self.transaction is None:
            raise ValueError("transaction is none, build transaction first.")

        unsigned_datas: list = []
        for signing_instruction in self.transaction["signing_instructions"]:
            unsigned_data = dict(datas=signing_instruction["sign_data"])
            if "pubkey" in signing_instruction and signing_instruction["pubkey"]:
                unsigned_data.setdefault("public_key", signing_instruction["pubkey"])
                if detail:
                    program = get_program(public_key=signing_instruction["pubkey"])
                    address = get_address(program=program, network=self.network)
                    unsigned_data.setdefault("program", program)
                    unsigned_data.setdefault("address", address)
                else:
                    unsigned_data.setdefault("network", self.network)
            else:
                if detail:
                    unsigned_data.setdefault("public_key", None)
                    unsigned_data.setdefault("program", None)
                    unsigned_data.setdefault("address", None)
                else:
                    unsigned_data.setdefault("network", self.network)
            if "derivation_path" in signing_instruction and signing_instruction["derivation_path"]:
                path = indexes_to_path(indexes=signing_instruction["derivation_path"])
                if detail:
                    unsigned_data.setdefault("indexes", signing_instruction["derivation_path"])
                unsigned_data.setdefault("path", path)
            else:
                if detail:
                    unsigned_data.setdefault("indexes", None)
                unsigned_data.setdefault("path", None)
            # Append unsigned datas
            unsigned_datas.append(unsigned_data)
        # Returning
        return unsigned_datas

    # Signing transaction using xprivate keys
    def sign(self, xprivate_key: str, account=1,
             change=False, address=1, path=None, indexes=None):
        """
        Bytom sign transaction.

        :param xprivate_key: bytom xprivate key.
        :type xprivate_key: str
        :param account: bytom derivation account, defaults to 1.
        :type account: int
        :param change: bytom derivation change, defaults to False.
        :type change: bool
        :param address: bytom derivation address, defaults to 1.
        :type address: int
        :param path: Bytom derivation path, default to None.
        :type path: str.
        :param indexes: Bytom derivation indexes, default to None.
        :type indexes: list.
        :returns: Transaction -- bytom transaction instance.

        >>> from pybytom.transaction import Transaction
        >>> transaction = Transaction(network="mainnet")
        >>> transaction.build_transaction("f0ed6ddd-9d6b-49fd-8866-a52d1083a13b", inputs=[...], outputs=[...])
        >>> transaction.sign(xprivate_key)
        <pybytom.transaction.transaction.Transaction object at 0x0409DAF0>
        """

        if not isinstance(xprivate_key, str):
            raise TypeError("invalid xprivate_key instance, only takes string type")
        if not isinstance(account, int):
            raise TypeError("invalid account instance, only takes integer type")
        if not isinstance(change, bool):
            raise TypeError("invalid change instance, only takes boolean type")
        if not isinstance(address, int):
            raise TypeError("invalid address instance, only takes integer type")
        if path and not isinstance(path, str):
            raise TypeError("invalid path instance, only takes string type")
        if indexes and not isinstance(indexes, list):
            raise TypeError("invalid indexes instance, only takes list type")

        wallet = Wallet(network=self.network)
        wallet.from_xprivate_key(xprivate_key=xprivate_key)
        for signing_instruction in self.unsigned_datas(detail=True):
            signed_data: list = []
            unsigned_datas = signing_instruction["datas"]
            if signing_instruction["path"]:
                wallet.from_path(signing_instruction["path"])
            elif not path and not indexes:
                wallet.from_path(f"m/44/153/{account}/{1 if change else 0}/{address}")
            elif path:
                wallet.from_path(path)
            elif indexes:
                wallet.from_indexes(indexes)
            for unsigned_data in unsigned_datas:
                signed_data.append(wallet.sign(unsigned_data))
            self._signatures.append(signed_data)
            wallet.clean_derivation()
        return self

    # Signed datas/signatures
    def signatures(self) -> list:
        """
        Bytom signed datas/signatures.

        :returns: list -- Bytom sign datas/signatures.

        >>> from pybytom.transaction import Transaction
        >>> transaction = Transaction(network="mainnet")
        >>> transaction.build_transaction("f0ed6ddd-9d6b-49fd-8866-a52d1083a13b", inputs=[...], outputs=[...])
        >>> transaction.sign(xprivate_key="3842e3fa2af2a687e7fd67655e7a02e85bbb4ca378d4338ff590dedc7ddff447797e1e781190835138c2d1a96d0e654b625a4c019cbc64f71100be7ad1b8d4ed")
        >>> transaction.signatures()
        [[...], ...]
        """

        return self._signatures


# Normal transaction
class NormalTransaction(Transaction):
    """
    Bytom Normal transaction class.

    :param network: Bytom network, defaults to solonet.
    :type network: str

    :returns:  NormalTransaction -- Bytom normal transaction instance.

    .. note::
        Bytom has only three networks, ``mainnet``. ``solonet`` and ``testnet``.
    """

    # Init normal transaction
    def __init__(self, network="solonet"):
        # Transaction
        super().__init__(network)

    # Building bytom normal transaction
    def build_transaction(self, guid: str, recipients: dict, asset: str,
                          fee=config["fee"], confirmations=config["confirmations"]):
        """
        Build bytom normal transaction.

        :param guid: Bytom sender blockcenter guid.
        :type guid: str
        :param recipients: Recipients bytom address and amount.
        :type recipients: dict
        :param asset: Bytom asset id, defaults to BTM asset.
        :type asset: str
        :param fee: Bytom transaction fee, defaults to 10000000.
        :type fee: int
        :param confirmations: Bytom transaction confirmations, defaults to 1.
        :type confirmations: int
        :returns: NormalTransaction -- Bytom normal transaction instance.

        >>> from pybytom.transaction import NormalTransaction
        >>> normal_transaction = NormalTransaction(network="mainnet")
        >>> normal_transaction.build_transaction("f0ed6ddd-9d6b-49fd-8866-a52d1083a13b", inputs=[...], outputs=[...])
        <pybytom.transaction.transaction.NormalTransaction object at 0x0409DAF0>
        """

        # Checking build transaction arguments instance
        if not isinstance(guid, str):
            raise TypeError("invalid guid instance, only takes bytom string type")
        if not isinstance(recipients, dict):
            raise TypeError("invalid recipients instance, only takes bytom dictionary type")
        if not isinstance(asset, str):
            raise TypeError("invalid asset instance, only takes string type")
        # Actions
        inputs, outputs = list(), list()
        # Input action
        inputs.append(
            spend_wallet_action(
                asset=asset,
                amount=sum(recipients.values())
            )
        )
        # Output action
        for address, amount in recipients.items():
            outputs.append(
                control_address_action(
                    asset=asset,
                    amount=amount,
                    address=address
                )
            )
        # Setting transaction fee and confirmations
        self._fee, self._confirmations = fee, confirmations
        # Transaction
        transaction = dict(
            guid=guid,
            inputs=inputs,
            outputs=outputs,
            fee=self._fee,
            confirmations=self._confirmations
        )
        # Building transaction
        self.transaction = build_transaction(
            transaction=transaction, network=self.network)
        return self

    # Transaction json
    def json(self) -> any:
        """
        Get bytom normal transaction json format.

        :returns: dict -- bytom normal transaction json format.

        >>> from pybytom.transaction import NormalTransaction
        >>> normal_transaction = NormalTransaction(network="mainnet")
        >>> normal_transaction.build_transaction("f0ed6ddd-9d6b-49fd-8866-a52d1083a13b", inputs=[...], outputs=[...])
        >>> normal_transaction.json()
        {"hash": "2993414225f65390220730d0c1a356c14e91bca76db112d37366df93e364a492", "status_fail": false, "size": 379, "submission_timestamp": 0, "memo": "", "inputs": [{"script": "00142cda4f99ea8112e6fa61cdd26157ed6dc408332a", "address": "bm1q9ndylx02syfwd7npehfxz4lddhzqsve2fu6vc7", "asset": "ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff", "amount": 2450000000, "type": "spend"}], "outputs": [{"utxo_id": "5edccebe497893c289121f9e365fdeb34c97008b9eb5a9960fe9541e7923aabc", "script": "01642091ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e220ac13c0bb1445423a641754182d53f0677cd4351a0e743e6f10b35122c3d7ea01202b9a5949f5546f63a253e41cda6bffdedb527288a7e24ed953f5c2680c70d6ff741f547a6416000000557aa888537a7cae7cac631f000000537acd9f6972ae7cac00c0", "address": "smart contract", "asset": "ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff", "amount": 1000, "type": "control"}, {"utxo_id": "f8cfbb692db1963be88b09c314adcc9e19d91c6c019aa556fb7cb76ba8ffa1fa", "script": "00142cda4f99ea8112e6fa61cdd26157ed6dc408332a", "address": "bm1q9ndylx02syfwd7npehfxz4lddhzqsve2fu6vc7", "asset": "ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff", "amount": 2439999000, "type": "control"}], "fee": 10000000, "balances": [{"asset": "ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff", "amount": "-10001000"}], "types": ["ordinary"]}
        """

        if self.transaction is None:
            raise ValueError("transaction is none, build transaction first.")
        # return decode_tx_raw(tx_raw=self.transaction["raw_transaction"])
        return self.transaction["tx"]


# Advanced transaction
class AdvancedTransaction(Transaction):
    """
    Bytom Advanced transaction class.

    :param network: Bytom network, defaults to solonet.
    :type network: str

    :returns:  AdvancedTransaction -- Bytom advanced transaction instance.

    .. note::
        Bytom has only three networks, ``mainnet``. ``solonet`` and ``testnet``.
    """

    # Init advanced transaction
    def __init__(self, network="solonet"):
        # Transaction
        super().__init__(network)

    # Building bytom advanced transaction
    def build_transaction(self, guid: str, inputs: list, outputs: list,
                          fee=config["fee"], confirmations=config["confirmations"]):
        """
        Build bytom advanced transaction.

        :param guid: Bytom blockcenter guid.
        :type guid: str
        :param inputs: Bytom transaction inputs.
        :type inputs: list
        :param outputs: Bytom transaction outputs.
        :type outputs: list
        :param fee: Bytom transaction fee, defaults to 10000000.
        :type fee: int
        :param confirmations: bytom transaction confirmations, defaults to 1.
        :type confirmations: int
        :returns:  AdvancedTransaction -- Bytom advanced transaction instance.

        >>> from pybytom.transaction import AdvancedTransaction
        >>> advanced_transaction = AdvancedTransaction(network="mainnet")
        >>> advanced_transaction.build_transaction("f0ed6ddd-9d6b-49fd-8866-a52d1083a13b", [[...], ...], outputs=[[...], ...], 10000000, 3)
        <pybytom.transaction.transaction.AdvancedTransaction object at 0x0409DAF0>
        """

        # Checking build transaction arguments instance
        if not isinstance(guid, str):
            raise TypeError("invalid guid instance, only takes string type")
        if not isinstance(inputs, list):
            raise TypeError("invalid inputs instance, only takes list type")
        if not isinstance(outputs, list):
            raise TypeError("invalid outputs instance, only takes list type")
        if not isinstance(fee, int):
            raise TypeError("invalid fee instance, only takes integer type")
        if not isinstance(confirmations, int):
            raise TypeError("invalid confirmations instance, only takes integer type")

        # Setting transaction fee and confirmations
        self._fee, self._confirmations = fee, confirmations
        # Transaction
        transaction = dict(
            guid=guid,
            inputs=inputs,
            outputs=outputs,
            fee=self._fee,
            confirmations=self._confirmations
        )
        # Building transaction
        self.transaction = build_transaction(
            transaction=transaction, network=self.network)
        return self



#!/usr/bin/env python
# coding=utf-8

from dataclasses import dataclass


@dataclass()
class ClientError(Exception):
    error_message: str
    error_detail: str = None

    def __str__(self):
        if self.error_detail:
            return f"{self.error_message}, {self.error_detail}"
        else:
            return f"{self.error_message}"


@dataclass()
class APIError(Exception):
    error_message: str
    status_code: int = None

    def __str__(self):
        if self.status_code:
            return f"({self.status_code}), {self.error_message}"
        else:
            return f"{self.error_message}"


@dataclass()
class InvalidURLError(Exception):
    error_message: str

    def __str__(self):
        return f"{self.error_message}"


@dataclass()
class NotFoundError(Exception):
    error_message: str
    error_detail: str = None

    def __str__(self):
        if self.error_detail:
            return f"{self.error_message}, {self.error_detail}"
        else:
            return f"{self.error_message}"


@dataclass()
class AddressError(Exception):
    error_message: str
    error_detail: str = None

    def __str__(self):
        if self.error_detail:
            return f"{self.error_message}, {self.error_detail}"
        return f"{self.error_message}"


@dataclass()
class BalanceError(Exception):
    error_message: str
    error_detail: str = None

    def __str__(self):
        if self.error_detail:
            return f"{self.error_message}, {self.error_detail}"
        return f"{self.error_message}"


@dataclass()
class NetworkError(Exception):
    error_message: str
    error_detail: str = None

    def __str__(self):
        if self.error_detail:
            return f"{self.error_message}, {self.error_detail}"
        return f"{self.error_message}"



#!/usr/bin/env python3

from binascii import hexlify, unhexlify
from mnemonic.mnemonic import Mnemonic

import os


def generate_entropy(strength=128):
    """
    Generate entropy hex string.

    :param strength: Entropy strength, default to 128.
    :type strength: str.
    :returns:  entropy -- Entropy hex string.

    >>> from pybytom.utils import generate_entropy
    >>> generate_entropy(strength=128)
    "ee535b143b0d9d1f87546f9df0d06b1a"
    """

    if strength not in [128, 160, 192, 224, 256]:
        raise ValueError(
            "Strength should be one of the following "
            "[128, 160, 192, 224, 256], but it is not (%d)."
            % strength
        )
    return hexlify(os.urandom(strength // 8)).decode()


def generate_mnemonic(language="english", strength=128):
    """
    Generate 12 word mnemonic.

    :param language: Mnemonic language, default to english.
    :type language: str.
    :param strength: Entropy strength, default to 128.
    :type strength: str.
    :returns:  mnemonic -- 12 word mnemonic.

    >>> from pybytom.utils import generate_mnemonic
    >>> generate_mnemonic(language="french")
    "sceptre capter séquence girafe absolu relatif fleur zoologie muscle sirop saboter parure"
    """

    if language and language not in ["english", "french", "italian", "japanese",
                                     "chinese_simplified", "chinese_traditional", "korean", "spanish"]:
        raise ValueError("invalid language, use only this options english, french, "
                         "italian, spanish, chinese_simplified, chinese_traditional, japanese & korean.")
    if strength not in [128, 160, 192, 224, 256]:
        raise ValueError(
            "Strength should be one of the following "
            "[128, 160, 192, 224, 256], but it is not (%d)."
            % strength
        )

    return Mnemonic(language=language).generate(strength=strength)


def is_mnemonic(mnemonic, language=None):
    """
    Check 12 word mnemonic is Valid.

    :param mnemonic: 12 word mnemonic.
    :type mnemonic: str.
    :param language: Mnemonic language, default to None.
    :type language: str.
    :returns:  mnemonic -- True/False.

    >>> from pybytom.utils import is_mnemonic
    >>> is_mnemonic("sceptre capter séquence girafe absolu relatif fleur zoologie muscle sirop saboter parure")
    True
    """

    if language and language not in ["english", "french", "italian", "japanese",
                                     "chinese_simplified", "chinese_traditional", "korean", "spanish"]:
        raise ValueError("invalid language, use only this options english, french, "
                         "italian, spanish, chinese_simplified, chinese_traditional, japanese & korean.")
    try:
        if language is None:
            for _language in ["english", "french", "italian",
                              "chinese_simplified", "chinese_traditional", "japanese", "korean", "spanish"]:
                valid = False
                if Mnemonic(language=_language).check(mnemonic=mnemonic) is True:
                    valid = True
                    break
            return valid
        else:
            return Mnemonic(language=language).check(mnemonic=mnemonic)
    except:
        return False


def get_mnemonic_language(mnemonic):
    """
    Get mnemonic language.

    :param mnemonic: 12 word mnemonic.
    :type mnemonic: str.
    :returns:  language -- Mnemonic language.

    >>> from pybytom.utils import get_mnemonic_language
    >>> get_mnemonic_language("sceptre capter séquence girafe absolu relatif fleur zoologie muscle sirop saboter parure")
    "french"
    """

    if not is_mnemonic(mnemonic=mnemonic):
        raise ValueError("invalid 12 word mnemonic.")

    language = None
    for _language in ["english", "french", "italian",
                      "chinese_simplified", "chinese_traditional", "japanese", "korean", "spanish"]:
        if Mnemonic(language=_language).check(mnemonic=mnemonic) is True:
            language = _language
            break
    return language



#!/usr/bin/env python3

import hashlib

from .utils import get_expanded_private_key, _verify, get_public_key, \
    sc_reduce32, decodeint, encodepoint, scalarmultbase, get_xpublic_key, sc_muladd


def sign(private_key, message):
    """
    Sign bytom message data by private key.

    :param private_key: Bytom private key.
    :type private_key: str.
    :param message: Message data.
    :type message: str.
    :return: str -- bytom signed message or signature.

    >>> from pybytom.signature import sign
    >>> sign(bytom_private_key, message_data)
    "..."
    """

    private = get_expanded_private_key(private_key)
    private_bytes = bytes.fromhex(private)
    message_bytes = bytes.fromhex(message)
    data_bytes = private_bytes[32:64] + message_bytes

    message_digest = hashlib.sha512(data_bytes).digest()
    message_digest = sc_reduce32(message_digest.hex().encode())
    message_digest = bytes.fromhex(message_digest.decode())
    message_digest_reduced = message_digest[0:32]

    scalar = decodeint(message_digest_reduced)
    encoded_r = encodepoint(scalarmultbase(scalar))
    public = get_xpublic_key(private)
    public_bytes = bytes.fromhex(public)
    hram_digest_data = encoded_r + public_bytes[:32] + message_bytes

    hram_digest = hashlib.sha512(hram_digest_data).digest()
    hram_digest = sc_reduce32(hram_digest.hex().encode())
    hram_digest = bytes.fromhex(hram_digest.decode())
    hram_digest_reduced = hram_digest[0:32]

    sk = private_bytes[:32]
    s = sc_muladd(hram_digest_reduced.hex().encode(), sk.hex().encode(), message_digest_reduced.hex().encode())
    s = bytes.fromhex(s.decode())

    signature_bytes = encoded_r + s
    signature = signature_bytes.hex()
    return signature


def verify(public_key, message, signature):
    """
    Verify bytom signature by public key.

    :param public_key: Bytom public key.
    :type public_key: str.
    :param message: Message data.
    :type message: str.
    :param signature: Signed message data.
    :type signature: str.
    :return: bool -- verified signature.

    >>> from pybytom.signature import verify
    >>> verify(bytom_public_key, message_data, bytom_signature)
    True
    """

    return _verify(get_public_key(public_key), signature, message)



#!/usr/bin/env python3

from ..rpc import get_transaction


def find_contract_utxo_id(transaction_id, network):
    """
    Find smart contract UTXO id.

    :param transaction_id: Bytom transaction id or hash.
    :type transaction_id: str
    :param network: Bytom network.
    :type network: str
    :returns: str -- UTXO id.

    >>> from pybytom.transaction.tools import find_contract_utxo_id
    >>> find_contract_utxo_id("e4d4fab70a41536cf298d6f261c0a1ac169a45be47583f7240115c9059cd0d03", "mainnet")
    "169a45be47583f7240115c9059cd0d03e4d4fab70a41536cf298d6f261c0a1ac"
    """

    utxo_id = None
    contract_transaction = get_transaction(
        transaction_id=transaction_id, network=network)
    contract_outputs = contract_transaction["outputs"]
    for contract_output in contract_outputs:
        if contract_output["address"] == "smart contract":
            utxo_id = contract_output["utxo_id"]
            break
    return utxo_id


def spend_utxo_action(utxo):
    """
    Get spend UTXO action

    :param utxo: Bytom utxo id.
    :type utxo: str
    :returns: dict -- Bytom spend utxo action.

    >>> from pybytom.transaction.tools import spend_utxo_action
    >>> spend_utxo_action(bytom_utxo_id)
    {'type': 'spend_utxo, 'output_id': '...'}
    """

    return dict(type=str("spend_utxo"), output_id=utxo)


def contract_arguments(amount, address, value=None):
    """
    Get contract arguments.

    :param amount: Bytom amount.
    :type amount: int
    :param address: Bytom address.
    :type address: str
    :param value: value, default to None.
    :type value: str
    :returns: list -- Bytom contract arguments.

    >>> from pybytom.transaction.tools import contract_arguments
    >>> contract_arguments(bytom_amount, bytom_address)
    [{'integer': 100}, {'address': '...'}, {'data': ''}]
    """

    return [
        dict(type=str("integer"), value=amount),
        dict(type=str("address"), value=address),
        dict(type=str("data"), value=str() if value is None else value)
    ]


def spend_wallet_action(amount, asset):
    """
    Get spend wallet action.

    :param amount: Bytom amount.
    :type amount: int
    :param asset: Bytom asset.
    :type asset: str
    :returns: dict -- Bytom spend wallet action.

    >>> from pybytom.transaction.tools import spend_wallet_action
    >>> spend_wallet_action(bytom_amount, bytom_asset)
    {'type': 'spend_wallet', 'amount': 100, 'asset': '...'}
    """

    return dict(
        type=str("spend_wallet"),
        amount=amount, asset=asset
    )


def spend_account_action(account, amount, asset):
    """
    Get spend account action.

    :param account: Bytom account.
    :type account: str
    :param amount: Bytom amount.
    :type amount: int
    :param asset: Bytom asset.
    :type asset: str
    :returns: dict -- Bytom spend account action.

    >>> from pybytom.transaction.tools import spend_account_action
    >>> spend_account_action(bytom_account, bytom_amount, bytom_asset)
    {type: 'spend_account', account='...', amount=1000, asset='...'}
    """

    return dict(type=str("spend_account"),
                account=account, amount=amount, asset=asset)


def control_program_action(amount, asset, control_program):
    """
    Get control program action.

    :param amount: Bytom amount.
    :type amount: int
    :param asset: Bytom asset.
    :type asset: str
    :param control_program: Bytom control program.
    :type control_program: str
    :returns: dict -- Bytom control program action.

    >>> from pybytom.transaction.tools import control_program_action
    >>> control_program_action(bytom_amount, bytom_asset, bytom_control_program)
    {'type': 'control_program', 'amount': 100, 'asset': '...', 'control_program': '...'}
    """

    return dict(
        type=str("control_program"),
        amount=amount, asset=asset,
        control_program=control_program
    )


def control_address_action(amount, asset, address):
    """
    Get control address action.

    :param amount: Bytom amount.
    :type amount: int
    :param asset: Bytom asset.
    :type asset: str
    :param address: Bytom address.
    :type address: str
    :returns: dict -- Bytom control address action.

    >>> from pybytom.transaction.tools import control_address_action
    >>> control_address_action(bytom_amount, bytom_asset, bytom_address)
    {'type': 'control_address', 'amount': 10000000, 'asset': '...', 'address': '...'}
    """

    return dict(
        type=str("control_address"),
        amount=amount, asset=asset, address=address
    )



#!/usr/bin/env python3

import requests
import json

from .exceptions import APIError

# Request headers
headers = dict()
# headers.setdefault("Content-Type", "application/json")

# Bytom configuration
config = {
    "mainnet": {
        "bytom": "http://localhost:9888",
        "blockmeta": "https://blockmeta.com/api/v3",
        "blockcenter": "https://bcapi.bystack.com/api/v2/btm"
    },
    "solonet": {
        "bytom": "http://localhost:9888",
        "blockmeta": None,
        "blockcenter": None
    },
    "testnet": {
        "bytom": "http://localhost:9888",
        "blockmeta": None,
        "blockcenter": None
    },
    "network": "solonet",  # mainnet, solonet & testnet
    "timeout": 60,
    "BTM_ASSET": "ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff",
    "fee": 10_000_000,  # 0.1 BTM
    "confirmations": 1
}


# Get balance by address
def get_balance(address, asset=config["BTM_ASSET"], network=config["network"], limit=1, page=1, timeout=config["timeout"]):
    """
    Get Bytom balance.

    :param address: Bytom address.
    :type address: str
    :param asset: Bytom asset, default to BTM asset.
    :type asset: str
    :param network: Bytom network, defaults to solonet.
    :type network: str
    :param limit: Bytom limit, defaults to 1.
    :type limit: str
    :param page: Bytom network, defaults to 1.
    :type page: str
    :param timeout: request timeout, default to 15.
    :type timeout: int
    :returns: int -- Bytom balance.

    >>> from pybytom.rpc import get_balance
    >>> get_balance(bytom_address, "mainnet")
    25800000
    """

    parameter = dict(limit=limit, page=page)
    url = f"{config[network]['blockmeta']}/address/{address}/asset"
    response = requests.get(url=url, params=parameter,
                            headers=headers, timeout=timeout)
    if response.status_code == 204:
        return 0
    for _asset in response.json():
        if asset == _asset["asset_id"]:
            return _asset["balance"]
    return None


# Create account in blockcenter
def account_create(xpublic_key, label="1st address", email=None,
                   network=config["network"], timeout=config["timeout"]):
    """
    Create account in blockcenter.

    :param xpublic_key: Bytom xpublic key.
    :type xpublic_key: str
    :param label: Bytom limit, defaults to 1st address.
    :type label: str
    :param email: email address, defaults to None.
    :type email: str
    :param network: Bytom network, defaults to solonet.
    :type network: str
    :param timeout: request timeout, default to 15.
    :type timeout: int
    :returns: dict -- Bytom blockcenter guid, address and label.

    >>> from pybytom.rpc import account_create
    >>> account_create(xpublic_key, "mainnet")
    {"guid": "f0ed6ddd-9d6b-49fd-8866-a52d1083a13b", "address": "bm1q9ndylx02syfwd7npehfxz4lddhzqsve2fu6vc7", "label": "1st address"}
    """

    url = str(config[network]["blockcenter"]) + "/account/create"
    data = dict(pubkey=xpublic_key, label=label, email=email)
    response = requests.post(url=url, data=json.dumps(data),
                             headers=headers, timeout=timeout)
    if response.status_code == 200 and response.json()["code"] == 300:
        raise APIError(response.json()["code"], response.json()["msg"])
    return response.json()["result"]["data"]


# List addresses from blockcenter
def list_address(guid, limit=10, network=config["network"], timeout=config["timeout"]):
    """
    List address from blockcenter.

    :param guid: Bytom blockcenter guid.
    :type guid: str
    :param limit: blockcenter limit default to 10.
    :type limit: int
    :param network: Bytom network, defaults to solonet.
    :type network: str
    :param timeout: request timeout, default to 15.
    :type timeout: int
    :returns: list -- Bytom blockcenter list of addresses.

    >>> from pybytom.rpc import list_address
    >>> list_address(guid, 5 "mainnet")
    [{"guid": "f0ed6ddd-9d6b-49fd-8866-a52d1083a13b", "address": "bm1q9ndylx02syfwd7npehfxz4lddhzqsve2fu6vc7", "label": "1st address", "balances": [{"asset": "f37dea62efd2965174b84bbb59a0bd0a671cf5fb2857303ffd77c1b482b84bdf", "balance": "100000000000", "total_received": "100000000000", "total_sent": "0", "decimals": 8, "alias": "Asset", "icon": "", "name": "f37dea62efd2965174b84bbb59a0bd0a671cf5fb2857303ffd77c1b482b84bdf", "symbol": "Asset", "in_usd": "0.00", "in_cny": "0.00", "in_btc": "0.000000"}, {"asset": "ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff", "balance": "2450000000", "total_received": "4950000000", "total_sent": "2500000000", "decimals": 8, "alias": "btm", "icon": "", "name": "BTM", "symbol": "BTM", "in_usd": "2.90", "in_cny": "20.58", "in_btc": "0.000283"}]}]
    """

    url = str(config[network]["blockcenter"]) + "/account/list-address"
    response = requests.post(url=url, data=json.dumps(dict(guid=guid)),
                             params=dict(limit=limit), headers=headers, timeout=timeout)
    if response.status_code == 200 and response.json()["code"] == 300:
        raise APIError(response.json()["code"], response.json()["msg"])
    return response.json()["result"]["data"]


# Build transaction in blockcenter
def build_transaction(transaction, network=config["network"], timeout=config["timeout"]):
    """
    Build Bytom transaction in blockcenter.

    :param transaction: Bytom transaction.
    :type transaction: dict
    :param network: Bytom network, defaults to solonet.
    :type network: str
    :param timeout: request timeout, default to 15.
    :type timeout: int
    :returns: dict -- Bytom built transaction.

    >>> from pybytom.rpc import build_transaction
    >>> build_transaction(transaction, "mainnet")
    {"transaction": {"hash": "2993414225f65390220730d0c1a356c14e91bca76db112d37366df93e364a492", "status_fail": false, "size": 379, "submission_timestamp": 0, "memo": "", "inputs": [{"script": "00142cda4f99ea8112e6fa61cdd26157ed6dc408332a", "address": "bm1q9ndylx02syfwd7npehfxz4lddhzqsve2fu6vc7", "asset": "ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff", "amount": 2450000000, "type": "spend"}], "outputs": [{"utransactiono_id": "5edccebe497893c289121f9e365fdeb34c97008b9eb5a9960fe9541e7923aabc", "script": "01642091ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e220ac13c0bb1445423a641754182d53f0677cd4351a0e743e6f10b35122c3d7ea01202b9a5949f5546f63a253e41cda6bffdedb527288a7e24ed953f5c2680c70d6ff741f547a6416000000557aa888537a7cae7cac631f000000537acd9f6972ae7cac00c0", "address": "smart contract", "asset": "ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff", "amount": 1000, "type": "control"}, {"utransactiono_id": "f8cfbb692db1963be88b09c314adcc9e19d91c6c019aa556fb7cb76ba8ffa1fa", "script": "00142cda4f99ea8112e6fa61cdd26157ed6dc408332a", "address": "bm1q9ndylx02syfwd7npehfxz4lddhzqsve2fu6vc7", "asset": "ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff", "amount": 2439999000, "type": "control"}], "fee": 10000000, "balances": [{"asset": "ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff", "amount": "-10001000"}], "types": ["ordinary"]}, "raw_transaction": "070100010160015e7f2d7ecec3f61d30d0b2968973a3ac8448f0599ea20dce883b48c903c4d6e87fffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff8091a0900901011600142cda4f99ea8112e6fa61cdd26157ed6dc408332a22012091ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e20201ad01ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffe80701880101642091ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e220ac13c0bb1445423a641754182d53f0677cd4351a0e743e6f10b35122c3d7ea01202b9a5949f5546f63a253e41cda6bffdedb527288a7e24ed953f5c2680c70d6ff741f547a6416000000557aa888537a7cae7cac631f000000537acd9f6972ae7cac00c000013dffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff98dcbd8b09011600142cda4f99ea8112e6fa61cdd26157ed6dc408332a00", "signing_instructions": [{"derivation_path": ["2c000000", "99000000", "01000000", "00000000", "01000000"], "sign_data": ["37727d44af9801e9723eb325592f4d55cc8d7e3815b1d663d61b7f1af9fc13a7"], "pubkey": "91ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e2"}], "fee": 10000000}
    """

    url = str(config[network]["blockcenter"]) + "/merchant/build-transaction"
    response = requests.post(url=url, data=json.dumps(transaction),
                             headers=headers, timeout=timeout)
    if response.status_code == 200 and response.json()["code"] == 300:
        raise APIError(response.json()["code"], response.json()["msg"])
    elif response.status_code == 200 and response.json()["code"] == 503:
        raise APIError(response.json()["code"], response.json()["msg"])
    return response.json()["result"]["data"]


# Get transaction from blockcenter
def get_transaction(transaction_id, network=config["network"], timeout=config["timeout"]):
    """
    Get Bytom transaction detail.

    :param transaction_id: Bytom transaction id.
    :type transaction_id: str
    :param network: Bytom network, defaults to solonet.
    :type network: str
    :param timeout: request timeout, default to 15.
    :type timeout: int
    :returns: dict -- Bytom built transaction.

    >>> from pybytom.rpc import get_transaction
    >>> get_transaction(transaction_id, "mainnet")
    {"transaction": {"hash": "2993414225f65390220730d0c1a356c14e91bca76db112d37366df93e364a492", "status_fail": false, "size": 379, "submission_timestamp": 0, "memo": "", "inputs": [{"script": "00142cda4f99ea8112e6fa61cdd26157ed6dc408332a", "address": "bm1q9ndylx02syfwd7npehfxz4lddhzqsve2fu6vc7", "asset": "ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff", "amount": 2450000000, "type": "spend"}], "outputs": [{"utransactiono_id": "5edccebe497893c289121f9e365fdeb34c97008b9eb5a9960fe9541e7923aabc", "script": "01642091ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e220ac13c0bb1445423a641754182d53f0677cd4351a0e743e6f10b35122c3d7ea01202b9a5949f5546f63a253e41cda6bffdedb527288a7e24ed953f5c2680c70d6ff741f547a6416000000557aa888537a7cae7cac631f000000537acd9f6972ae7cac00c0", "address": "smart contract", "asset": "ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff", "amount": 1000, "type": "control"}, {"utransactiono_id": "f8cfbb692db1963be88b09c314adcc9e19d91c6c019aa556fb7cb76ba8ffa1fa", "script": "00142cda4f99ea8112e6fa61cdd26157ed6dc408332a", "address": "bm1q9ndylx02syfwd7npehfxz4lddhzqsve2fu6vc7", "asset": "ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff", "amount": 2439999000, "type": "control"}], "fee": 10000000, "balances": [{"asset": "ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff", "amount": "-10001000"}], "types": ["ordinary"]}, "raw_transaction": "070100010160015e7f2d7ecec3f61d30d0b2968973a3ac8448f0599ea20dce883b48c903c4d6e87fffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff8091a0900901011600142cda4f99ea8112e6fa61cdd26157ed6dc408332a22012091ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e20201ad01ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffe80701880101642091ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e220ac13c0bb1445423a641754182d53f0677cd4351a0e743e6f10b35122c3d7ea01202b9a5949f5546f63a253e41cda6bffdedb527288a7e24ed953f5c2680c70d6ff741f547a6416000000557aa888537a7cae7cac631f000000537acd9f6972ae7cac00c000013dffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff98dcbd8b09011600142cda4f99ea8112e6fa61cdd26157ed6dc408332a00", "signing_instructions": [{"derivation_path": ["2c000000", "99000000", "01000000", "00000000", "01000000"], "sign_data": ["37727d44af9801e9723eb325592f4d55cc8d7e3815b1d663d61b7f1af9fc13a7"], "pubkey": "91ff7f525ff40874c4f47f0cab42e46e3bf53adad59adef9558ad1b6448f22e2"}], "fee": 10000000}
    """

    url = str(config[network]["blockcenter"]) + "/merchant/get-transaction"
    response = requests.post(url=url, data=json.dumps(dict(tx_id=transaction_id)),
                             headers=headers, timeout=timeout)
    if response.status_code == 200 and response.json()["code"] == 300:
        raise APIError(response.json()["msg"], response.json()["code"])
    return response.json()["result"]["data"]


# Submit transaction raw from blockcenter
def submit_transaction_raw(guid, transaction_raw, signatures,
                           network, memo="mock", timeout=config["timeout"]):
    """
     Submit transaction raw to Bytom blockchain.

    :param guid: Bytom blockcenter id.
    :type guid: str
    :param transaction_raw: Bytom transaction raw.
    :type transaction_raw: str
    :param signatures: Bytom signed datas.
    :type signatures: list
    :param network: Bytom network, defaults to solonet.
    :type network: str
    :param memo: memo, defaults to mock.
    :type memo: str
    :param timeout: request timeout, default to 15.
    :type timeout: int
    :returns: dict -- Bytom transaction id, fee, type and date.

    >>> from pybytom.rpc import submit_transaction_raw
    >>> submit_transaction_raw("guid", transaction_raw, [[...], [...]], "mainent")
    {...}
    """
    if not isinstance(signatures, list):
        raise TypeError("signatures must be list format")
    url = str(config[network]["blockcenter"]) + "/merchant/submit-payment"
    data = dict(guid=guid, raw_transaction=transaction_raw, signatures=signatures, memo=memo)
    response = requests.post(url=url, data=json.dumps(data),
                             headers=headers, timeout=timeout)
    if response.json()["code"] != 200:
        raise APIError(response.json()["code"], response.json()["msg"])
    return response.json()["result"]["data"]["transaction_hash"]


# Decode transaction raw
def decode_transaction_raw(transaction_raw, network=config["network"], timeout=config["timeout"]):
    """
    Get decoded transaction raw.

    :param transaction_raw: Bytom transaction raw.
    :type transaction_raw: str
    :param network: Bytom network, defaults to solonet.
    :type network: str
    :param timeout: request timeout, default to 15.
    :type timeout: int
    :returns: dict -- Bytom decoded transaction raw.

    >>> from pybytom.rpc import decode_transaction_raw
    >>> decode_transaction_raw(transaction_raw, "testnet")
    {...}
    """

    url = str(config[network]["bytom"]) + "/decode-raw-transaction"
    response = requests.post(url=url, data=json.dumps(dict(raw_transaction=transaction_raw)),
                             headers=headers, timeout=timeout)
    if response.status_code == 400:
        raise APIError(response.json()["code"], response.json()["msg"])
    return response.json()["data"]



#!/usr/bin/env python3

from .wallet import Wallet
from .transaction import (
    Transaction, NormalTransaction, AdvancedTransaction
)
from .signature import sign, verify


__all__ = [
    "Wallet",
    "Transaction",
    "NormalTransaction",
    "AdvancedTransaction",
    "sign",
    "verify",
]



#!/usr/bin/env python3

from binascii import hexlify, unhexlify

import ed25519
import hmac

from ..libs.ed25519 import *


L = 2 ** 252 + 27742317777372353535851937790883648493


def get_xpublic_key(xprivate):
    xprivate_bytes = bytes.fromhex(xprivate)
    scalar = decodeint(xprivate_bytes[:len(xprivate_bytes)//2])
    buf = encodepoint(scalarmultbase(scalar))
    xpublic = buf + xprivate_bytes[len(xprivate_bytes)//2:]
    xpublic = xpublic.hex()
    return xpublic


def get_expanded_private_key(xprivate):
    hc_hexstr = hmac.HMAC(b"Expand", bytes.fromhex(xprivate), digestmod=hashlib.sha512).hexdigest()
    expanded_private_key_hexstr = xprivate[:64] + hc_hexstr[64:]
    return expanded_private_key_hexstr


def get_public_key(xpublic):
    public_key_hexstr = xpublic[:64]
    return public_key_hexstr


def _verify(public_key_str, signature_str, message_str):
    result = False
    verifying_key = ed25519.VerifyingKey(public_key_str.encode(), encoding='hex')
    try:
        verifying_key.verify(signature_str.encode(), bytes.fromhex(message_str), encoding='hex')
        result = True
    except ed25519.BadSignatureError:
        result = False
    return result


def byte2int(_b):
    return _b


def hex2int(_hex):
    unhex = unhexlify(_hex)
    s = 0
    for i in range(len(unhex)):
        s += 256 ** i * byte2int(unhex[i])
    return s


def int2hex(_int):
    return hexlify(encodeint(_int))


def sc_reduce32(inputs):
    _int = hex2int(inputs)
    modulo = _int % L
    return int2hex(modulo)


def sc_muladd(a, _b, c):
    a_int = hex2int(a)
    b_int = hex2int(_b)
    c_int = hex2int(c)

    s = a_int * b_int + c_int
    modulo = s % L
    return int2hex(modulo)
